/**************************************************************************\
*								 3DVECTOR.H								   *
*																		   *
*			 Data types and functions for 3D perspective drawing		   *
*						  Vectors, points and view						   *
*																		   *
*								Version 1.3b							   *
*																		   *
*							  by Stewart Gordon							   *
*							 July 1998-June 1999						   *
\**************************************************************************/
#ifndef __3DVECTOR_H
#define __3DVECTOR_H

#ifndef __GRAPHICS_H
#include <graphics.h>
#endif

#ifndef __STDIO_H
#include <stdio.h>
#endif

#ifndef __MISCIDS_H
#include <miscids.h>
#endif

#ifndef __POLYBASE_H
#include "polybase.h"
#endif

#ifndef __SCRPOLY_H
#include "scrpoly.h"
#endif

#ifndef __PALETTE_H
#include "palette.h"
#endif

#define rad(x)		((x)*PI/180)
#define deg(x)		((x)*180/PI)

// DECLARATIONS
class view;
extern view *curview;

// DATA TYPES

// Enum orderrel: position in stacking order relative to another shape
enum orderrel {
	BEHIND	= -1,	// second behind first
	EQUAL	=  0,	// coplanar or intersecting
	INFRONT	=  1,	// second in front of first
};

// Class vectorbase: base for 3D vector and point
class vectorbase {
public:
	float x, y, z;

	vectorbase() : x(0), y(0), z(0) {}
	vectorbase(float X, float Y, float Z): x(X), y(Y), z(Z) {}

	vectorbase operator+(vectorbase) const;
	vectorbase operator-(vectorbase) const;
	vectorbase operator-() const;
	vectorbase operator*(vectorbase) const;
	float operator%(vectorbase) const;

	vectorbase operator*(float a) const {
		return vectorbase(x * a, y * a, z * a);
	}

	vectorbase operator/(float a) const { return operator*(1 / a); }

	vectorbase operator+=(vectorbase a) { return *this = operator+(a); }
	vectorbase operator-=(vectorbase a) { return *this = operator-(a); }
	vectorbase operator*=(float a) { return *this = operator*(a); }
	vectorbase operator/=(float a) { return *this = operator/(a); }

	float lensq() const;
};

// Class vector: 3D vector
class vector : public vectorbase {
public:
	vector() : vectorbase() {}
	vector(vectorbase v) : vectorbase(v) {}
	vector(float X, float Y, float Z): vectorbase(X, Y, Z) {}
};

// Class relvector: apparent position of a vector in a particular view
class relvector: public vectorbase {
public:
	relvector() : vectorbase() {}
    relvector(vectorbase v) : vectorbase(v) {}
	relvector(float X, float Y, float Z) : vectorbase(X, Y, Z) {}
	relvector(vector, const view& = *curview);

	/*relvector operator+(relvector a) const { return vector::operator+(a); }
	relvector operator-(relvector a) const { return vector::operator-(a); }*/
	relvector operator-() const { return vectorbase::operator-(); }
	//relvector operator*(relvector a) const { return vector::operator*(a); }
};

// Class point: point in 3D space
class point: public vectorbase {
public:
	point() : vectorbase() {}
	point(vectorbase v) : vectorbase(v) {}
	point(float X, float Y, float Z) : vectorbase(X, Y, Z) {}
	float screenx(const view& = *curview) const;
	float screeny(const view& = *curview) const;
};

// Class relpoint: apparent position of a point in a particular view
class relpoint: public vectorbase {
public:
	relpoint() : vectorbase() {}
	relpoint(vectorbase v) : vectorbase(v) {}
	relpoint(float X, float Y, float Z): vectorbase(X, Y, Z) {}
	//relpoint(vector, const view& = *curview);
	relpoint(point, const view& = *curview);

	relpoint operator+(relvector a) const;
	relpoint operator-(relvector a) const;
	relvector operator-(relpoint a) const;

	float screenx(const view& = *curview) const;
	float screeny(const view& = *curview) const;
};

// Enum viewtype: type of view
enum viewtype { WIREFRAME, NORMAL, LIGHTSHADE };

// Class view: view of 3D space
class view {
public:
	point eye;			// coordinates of viewer's eye
	int vpx, vpy;		// coordinates on screen of vanishing point
	int alt, azim;		// altitude (vert angle) and azimuth (horz angle)
	int xscale, yscale;	// scaling factor
	viewtype vtype;		// wireframe, normal or light-shaded
	vector light;		// light source
	rgb lightcolour;	// directional light colour
	rgb ambcolour;		// ambient light colour
	palette *pal;		// dynamic colour palette

	view() : eye(0, 0, 0), vpx(320), vpy(240), alt(0), azim(0), xscale(200),
	  yscale(200), vtype(NORMAL), light(5, -4, 3), lightcolour(1, 1, 1),
	  pal(NULL) {}

	view(float x, float y, float z, int vx = 320, int vy = 240, int al = 0,
		int azm = 0, int xscl = 200, int yscl = 0, viewtype vt = NORMAL,
		float lx = 5, float ly = -4, float lz = 3)
	  : eye(x, y, z), vpx(vx), vpy(vy), alt(al), azim(azm), xscale(xscl),
		yscale(yscl == 0 ? xscl : yscl), vtype(vt), light(lx, ly, lz),
		lightcolour(.67, .67, .67), ambcolour(.33, .33, .33), pal(NULL) {}

	~view() { if(pal != NULL) delete pal; }

	const view& operator=(const view& vw);

	void fwd(float);
	void bwd(float d) { fwd(-d); }
	void fwd3(float);
	void bwd3(float d) { fwd3(-d); }

	void initpal(int, int);

	int palcolour(int red, int green, int blue) const {
		return pal->palcolour(red, green, blue);
	}
	int palcolour(rgb col) const { return pal->palcolour(col); }
};

#endif
